/*
    Copyright (C) 2001-2004 Paul Davis, Tilman Linneweh

    Generic version, overridden by OS-specific definition when needed.
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, write to the Free Software 
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/
#ifndef __jack_time_h__
#define __jack_time_h__

#include <jack/types.h>

extern jack_time_t jack_get_microseconds_from_system (void);

static inline jack_time_t
jack_get_microseconds (void) {
	return jack_get_microseconds_from_system ();
}

typedef jack_time_t (*jack_get_microseconds_t)(void);
static inline  jack_get_microseconds_t jack_get_microseconds_pointer(void)
{
	return jack_get_microseconds_from_system;
}

#endif /* __jack_time_h__ */
